#
#  tables.py
#

from os.path import join, dirname, splitext
import json
from typing import Any, Literal, Optional, List, Dict, Union, cast
from collections import defaultdict

import pandas as pd
import requests

from . import variables
from .meta import VariableMeta, TableMeta
from .frames import repack_frame

SCHEMA = json.load(open(join(dirname(__file__), "schemas", "table.json")))
METADATA_FIELDS = list(SCHEMA["properties"])


class Table(pd.DataFrame):
    # metdata about the entire table
    metadata: TableMeta

    # metadata about individual columns
    # NOTE: the name _fields is also on the Variable class, pandas will propagate this to
    #       any slices, which is how they get access to their metadata
    _fields: Dict[str, VariableMeta]

    # propagate all these fields on every slice or copy
    _metadata = ["metadata", "_fields"]

    # slicing and copying creates tables
    @property
    def _constructor(self) -> type:
        return Table

    @property
    def _constructor_sliced(self) -> Any:
        return variables.Variable

    def __init__(
        self, *args: Any, metadata: Optional[TableMeta] = None, **kwargs: Any
    ) -> None:
        super().__init__(*args, **kwargs)

        # empty table metadata by default
        self.metadata = metadata or TableMeta()

        # all columns have empty metadata by default
        assert not hasattr(self, "_fields")
        self._fields = defaultdict(VariableMeta)

    @property
    def primary_key(self) -> List[str]:
        return [n for n in self.index.names if n]

    # Mypy complaints about this not matching the defintiion of NDFrame.to_csv but I don't understand why
    def to_csv(self, path: Any, **kwargs: Any) -> None:  # type: ignore
        """
        Save this table as a csv file plus accompanying JSON metadata file.
        If the table is stored at "mytable.csv", the metadata will be at
        "mytable.meta.json".
        """
        if not isinstance(path, str) or not path.endswith(".csv"):
            raise ValueError(f'filename must end in ".csv": {path}')

        df = pd.DataFrame(self)
        # if the dataframe uses the default index then we don't want to store it (would be a column of row numbers)
        save_index = self.primary_key != []
        df.to_csv(path, index=save_index, **kwargs)

        metadata_filename = splitext(path)[0] + ".meta.json"
        self._save_metadata(metadata_filename)

    def to_feather(
        self,
        path: Any,
        repack: bool = True,
        compression: Literal["zstd", "lz4", "uncompressed"] = "zstd",
        **kwargs: Any,
    ) -> None:
        """
        Save this table as a feather file plus accompanying JSON metadata file.
        If the table is stored at "mytable.feather", the metadata will be at
        "mytable.meta.json".
        """
        if not isinstance(path, str) or not path.endswith(".feather"):
            raise ValueError(f'filename must end in ".feather": {path}')

        # feather can't store the index
        df = pd.DataFrame(self)
        if self.primary_key:
            df = df.reset_index()

        if repack:
            # use smaller data types wherever possible
            repack_frame(df)

        df.to_feather(path, compression=compression, **kwargs)

        metadata_filename = splitext(path)[0] + ".meta.json"
        self._save_metadata(metadata_filename)

    def _save_metadata(self, filename: str) -> None:
        # write metadata
        with open(filename, "w") as ostream:
            metadata = self.metadata.to_dict()  # type: ignore
            metadata["primary_key"] = self.primary_key
            metadata["fields"] = {
                col: self._fields[col].to_dict() for col in self.all_columns
            }
            json.dump(metadata, ostream, indent=2)

    @classmethod
    def read_csv(cls, path: str) -> "Table":
        """
        Read the table from csv plus accompanying JSON sidecar.
        """
        if not path.endswith(".csv"):
            raise ValueError(f'filename must end in ".csv": {path}')

        # load the data
        df = Table(
            pd.read_csv(path, index_col=False, na_values=[""], keep_default_na=False)
        )

        # load the metadata
        metadata = cls._read_metadata(path)

        primary_key = metadata.pop("primary_key") if "primary_key" in metadata else []
        fields = metadata.pop("fields") if "fields" in metadata else {}

        df.metadata = TableMeta(**metadata)
        df._fields = defaultdict(
            VariableMeta, {k: VariableMeta.from_dict(v) for k, v in fields.items()}
        )

        if primary_key:
            df.set_index(primary_key, inplace=True)

        return df

    def set_index(  # type: ignore
        self,
        keys: Union[str, List[str]],
        inplace: bool = False,
        drop: bool = True,
        append: bool = False,
        verify_integrity: bool = False,
    ) -> Optional[pd.DataFrame]:
        if isinstance(keys, str):
            keys = [keys]

        if inplace:
            super().set_index(
                keys,
                inplace=True,
                drop=drop,
                append=append,
                verify_integrity=verify_integrity,
            )
            self.metadata.primary_key = keys
            return None

        t = super().set_index(
            keys,
            inplace=False,
            drop=drop,
            append=append,
            verify_integrity=verify_integrity,
        )
        t.metadata.primary_key = keys
        return t

    @classmethod
    def read_feather(cls, path: str) -> "Table":
        """
        Read the table from feather plus accompanying JSON sidecar.
        """
        if not path.endswith(".feather"):
            raise ValueError(f'filename must end in ".feather": {path}')

        # load the data
        df = Table(pd.read_feather(path))

        # load the metadata
        metadata = cls._read_metadata(path)

        primary_key = metadata.get("primary_key", [])
        fields = metadata.pop("fields") if "fields" in metadata else {}

        df.metadata = TableMeta.from_dict(metadata)
        df._fields = defaultdict(
            VariableMeta, {k: VariableMeta.from_dict(v) for k, v in fields.items()}
        )

        if primary_key:
            df.set_index(primary_key, inplace=True)

        return df

    @staticmethod
    def _read_metadata(data_path: str) -> Dict[str, Any]:
        metadata_path = splitext(data_path)[0] + ".meta.json"

        if metadata_path.startswith("http"):
            return cast(Dict[str, Any], requests.get(metadata_path).json())

        with open(metadata_path, "r") as istream:
            return cast(Dict[str, Any], json.load(istream))

    def __setitem__(self, key: Any, value: Any) -> Any:
        super().__setitem__(key, value)

        # propagate metadata when we add a series to a table
        if isinstance(key, str):
            if isinstance(value, variables.Variable):
                self._fields[key] = value.metadata
            else:
                self._fields[key] = VariableMeta()

    def equals_table(self, rhs: "Table") -> bool:
        return (
            isinstance(rhs, Table)
            and self.metadata == rhs.metadata
            and self.to_dict() == rhs.to_dict()
        )

    @property
    def all_columns(self) -> List[str]:
        "Return names of all columns in the dataset, including the index."
        combined: List[str] = filter(None, list(self.index.names) + list(self.columns))  # type: ignore
        return combined



#
#  meta.py
#
#  Metadata helpers.
#

from pathlib import Path
from typing import Optional, TypeVar, Dict, Any, List, Union
from dataclasses import dataclass, field
import json

from dataclasses_json import dataclass_json

T = TypeVar("T")


def pruned_json(cls: T) -> T:
    orig = cls.to_dict  # type: ignore

    # only keep non-null public variables
    cls.to_dict = lambda self: {  # type: ignore
        k: v
        for k, v in orig(self).items()
        if not k.startswith("_") and v not in [None, [], {}]
    }

    return cls


@pruned_json
@dataclass_json
@dataclass
class Source:
    name: Optional[str] = None
    description: Optional[str] = None
    url: Optional[str] = None
    source_data_url: Optional[str] = None
    owid_data_url: Optional[str] = None
    date_accessed: Optional[str] = None
    publication_date: Optional[str] = None
    publication_year: Optional[int] = None

    def to_dict(self) -> Dict[str, Any]:
        ...


@pruned_json
@dataclass_json
@dataclass
class License:
    name: Optional[str]
    url: Optional[str]

    def to_dict(self) -> Dict[str, Any]:
        ...


@pruned_json
@dataclass_json
@dataclass
class VariableMeta:
    title: Optional[str] = None
    description: Optional[str] = None
    sources: List[Source] = field(default_factory=list)
    licenses: List[Source] = field(default_factory=list)
    unit: Optional[str] = None
    short_unit: Optional[str] = None
    display: Optional[Dict[str, Any]] = None

    def to_dict(self) -> Dict[str, Any]:
        ...

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "VariableMeta":
        ...


@pruned_json
@dataclass_json
@dataclass
class DatasetMeta:
    """
    The metadata for this entire dataset kept in JSON (e.g. mydataset/index.json).

    The number of fields is limited, but should handle everything that we get from
    Walden. There is a lot more opportunity to store more metadata at the table and
    the variable level.
    """

    namespace: Optional[str] = None
    short_name: Optional[str] = None
    title: Optional[str] = None
    description: Optional[str] = None
    sources: List[Source] = field(default_factory=list)
    licenses: List[License] = field(default_factory=list)

    # an md5 checksum of the ingredients used to make this dataset
    source_checksum: Optional[str] = None

    def save(self, filename: Union[str, Path]) -> None:
        filename = Path(filename).as_posix()
        with open(filename, "w") as ostream:
            json.dump(self.to_dict(), ostream, indent=2)

    @classmethod
    def load(cls, filename: str) -> "DatasetMeta":
        with open(filename) as istream:
            return cls.from_dict(json.load(istream))

    def to_dict(self) -> Dict[str, Any]:
        ...

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "DatasetMeta":
        ...

    @property
    def version(self) -> Optional[str]:
        if len(self.sources) == 1:
            (source,) = self.sources
            if source.publication_date:
                return source.publication_date

            return str(source.publication_year)

        return None


@pruned_json
@dataclass_json
@dataclass
class TableMeta:
    # data about this table
    short_name: Optional[str] = None
    title: Optional[str] = None
    description: Optional[str] = None

    # a reference back to the dataset
    dataset: Optional[DatasetMeta] = field(compare=False, default=None)
    primary_key: List[str] = field(default_factory=list)

    @property
    def checked_name(self) -> str:
        if not self.short_name:
            raise Exception("table has no short_name")

        return self.short_name

    @staticmethod
    def from_dict(dict: Dict[str, Any]) -> "TableMeta":
        ...



#
#  frames.py
#  etl
#

from typing import Any, Dict, List, Optional, Union, cast

import numpy as np
import pandas as pd


def repack_frame(df: pd.DataFrame, remap: Optional[Dict[str, str]] = None) -> None:
    """
    Convert the DataFrame's columns to the most compact types possible.
    Rename columns if necessary during the repacking. The column renames
    work even if the column is part of the index.
    """
    if df.index.names != [None]:
        raise ValueError("repacking is lost for index columns")

    remap = remap or {}

    # unwind the primary key
    if len(df.index.names) == 1 and not df.index.names[0]:
        primary_key = []
    else:
        primary_key = cast(List[str], df.index.names)
        df.reset_index(inplace=True)

    # repack each column into the best dtype we can give it
    for col in df.columns:
        df[col] = repack_series(df[col])

    # remap all column names, including those in the primary key
    for from_, to_ in remap.items():
        if from_ in df.columns:
            df.rename(columns={from_: to_}, inplace=True)
    primary_key = [remap.get(k, k) for k in primary_key]

    assert all(df[col].dtype != "object" for col in df.columns)

    # set the primary key back again
    if primary_key:
        df.set_index(primary_key, inplace=True)


def repack_series(s: pd.Series) -> pd.Series:
    if s.dtype.name in ("Int64", "int64", "UInt64", "uint64"):
        return shrink_integer(s)

    if s.dtype.name in ("object", "float64", "Float64"):
        for strategy in [to_int, to_float, to_category]:
            try:
                return strategy(s)
            except (ValueError, TypeError):
                continue

    return s


def to_int(s: pd.Series) -> pd.Series:
    # values could be integers or strings
    def intify(v: Any) -> Union[int, None]:
        return int(v) if not pd.isnull(v) else None

    v = cast(pd.Series, s.apply(intify).astype("Int64"))

    if not series_eq(v, s, cast=float):
        raise ValueError()

    # it's an integer, now pack it smaller
    return shrink_integer(v)


def shrink_integer(s: pd.Series) -> pd.Series:
    """
    Take an Int64 series and make it as small as possible.
    """
    assert s.dtype.name in ("Int64", "int64", "UInt64", "uint64")

    if s.isnull().any():
        if s.min() < 0:
            series = ["Int32", "Int16", "Int8"]
        else:
            series = ["UInt32", "UInt16", "UInt8"]
    else:
        if s.min() < 0:
            series = ["int32", "int16", "int8"]
        else:
            series = ["uint32", "uint16", "uint8"]

    for dtype in series:
        v = s.astype(dtype)
        if not (v == s).all():
            break

        s = v

    return s


def to_float(s: pd.Series) -> pd.Series:
    options = ["float32", "float64"]
    for dtype in options:
        v = s.astype(dtype)

        if series_eq(s, v, float):
            return v

    raise ValueError()


def to_category(s: pd.Series) -> pd.Series:
    types = set(s.dropna().apply(type).unique())

    if types.difference({str, type(None)}):
        raise ValueError()

    return s.astype("category")


def series_eq(
    lhs: pd.Series, rhs: pd.Series, cast: Any, rtol: float = 1e-5, atol: float = 1e-8
) -> bool:
    """
    Check that series are equal, but unlike normal floating point checks where
    NaN != NaN, we want missing or null values to be reported as equal to each
    other.
    """
    if len(lhs) != len(rhs) or (lhs.isnull() != rhs.isnull()).any():
        return False

    lhs_values = lhs.dropna().apply(cast)
    rhs_values = rhs.dropna().apply(cast)
    return np.allclose(lhs_values, rhs_values, rtol=rtol, atol=atol)



#
#  catalog.py
#  owid-catalog-py
#

from pathlib import Path
from typing import Dict, Optional, Iterator, Union, Any, cast
import json
import heapq

import pandas as pd
import numpy as np
import requests

from .datasets import Dataset
from .tables import Table

# increment this on breaking changes to require clients to update
OWID_CATALOG_VERSION = 1

# location of the default remote catalog
OWID_CATALOG_URI = "https://catalog.ourworldindata.org/"

# global copy cached after first request
REMOTE_CATALOG: Optional["RemoteCatalog"] = None


class CatalogMixin:
    """
    Abstract data catalog API, encapsulates finding and loading data.
    """

    frame: "CatalogFrame"

    def find(
        self,
        table: Optional[str] = None,
        namespace: Optional[str] = None,
        dataset: Optional[str] = None,
    ) -> "CatalogFrame":
        criteria = np.ones(len(self.frame), dtype=bool)

        if table:
            criteria &= self.frame.table.apply(lambda t: table in t)

        if namespace:
            criteria &= self.frame.namespace == namespace

        if dataset:
            criteria &= self.frame.dataset == dataset

        matches = self.frame[criteria]
        if "checksum" in matches.columns:
            matches = matches.drop(columns=["checksum"])

        return cast(CatalogFrame, matches)

    def find_one(self, *args: Optional[str], **kwargs: Optional[str]) -> Table:
        return self.find(*args, **kwargs).load()


class LocalCatalog(CatalogMixin):
    """
    A data catalog that's on disk. On-disk catalogs do not need an index file, since
    you can simply walk the directory. However, they support a `reindex()` method
    which can create such an index.
    """

    path: Path
    frame: "CatalogFrame"

    def __init__(self, path: Union[str, Path]) -> None:
        self.path = Path(path)
        if self._catalog_file.exists():
            df = pd.read_feather(self._catalog_file.as_posix())
            self.frame = CatalogFrame(df)
            self.frame._base_uri = self.path.as_posix() + "/"
        else:
            # could take a while to generate if there are many datasets
            self.reindex()

        # ensure the frame knows where to load data from

    @property
    def _catalog_file(self) -> Path:
        return self.path / "catalog.feather"

    @property
    def _metadata_file(self) -> Path:
        return self.path / "catalog.meta.json"

    def iter_datasets(self) -> Iterator[Dataset]:
        to_search = [self.path]
        while to_search:
            dir = heapq.heappop(to_search)
            if (dir / "index.json").exists():
                yield Dataset(dir)
                continue

            for child in dir.iterdir():
                if child.is_dir():
                    heapq.heappush(to_search, child)

    def reindex(self) -> None:
        self._save_metadata({"format_version": OWID_CATALOG_VERSION})

        # walk the directory tree, generate a namespace/version/dataset/table frame
        # save it to feather
        frames = []
        for ds in self.iter_datasets():
            frames.append(ds.index(self.path))

        df = pd.concat(frames, ignore_index=True)

        keys = ["table", "dataset", "version", "namespace"]
        columns = keys + [c for c in df.columns if c not in keys]

        df.sort_values(keys, inplace=True)
        df = df[columns]
        df.reset_index(drop=True, inplace=True)
        df.to_feather(self._catalog_file)

        self.frame = CatalogFrame(df)
        self.frame._base_uri = self.path.as_posix() + "/"

    def _save_metadata(self, contents: Dict[str, Any]) -> None:
        with open(self._metadata_file, "w") as ostream:
            json.dump(contents, ostream, indent=2)


class RemoteCatalog(CatalogMixin):
    uri: str
    frame: "CatalogFrame"

    def __init__(self, uri: str = OWID_CATALOG_URI) -> None:
        self.uri = uri
        self.metadata = self._read_metadata(self.uri + "catalog.meta.json")
        if self.metadata["format_version"] > OWID_CATALOG_VERSION:
            raise PackageUpdateRequired(
                f"library supports api version {OWID_CATALOG_VERSION}, "
                f'but the remote catalog has version {self.metadata["version"]} '
                "-- please update"
            )

        self.frame = CatalogFrame(pd.read_feather(self.uri + "catalog.feather"))
        self.frame._base_uri = uri

    @property
    def datasets(self) -> pd.DataFrame:
        return self.frame[["namespace", "version", "dataset"]].drop_duplicates()

    @staticmethod
    def _read_metadata(uri: str) -> Dict[str, Any]:
        """
        Read the metadata JSON blob for this repo.
        """
        resp = requests.get(uri)
        resp.raise_for_status()
        return cast(Dict[str, Any], resp.json())


class CatalogFrame(pd.DataFrame):
    """
    DataFrame helper, meant only for displaying catalog results.
    """

    _base_uri: Optional[str] = None

    _metadata = ["_base_uri"]

    @property
    def _constructor(self) -> type:
        return CatalogFrame

    @property
    def _constructor_sliced(self) -> Any:
        # ensure that when we pick a series we still have the URI
        def build(*args: Any, **kwargs: Any) -> Any:
            c = CatalogSeries(*args, **kwargs)
            c._base_uri = self._base_uri
            return c

        return build

    def load(self) -> Table:
        if len(self) == 1:
            return self.iloc[0].load()  # type: ignore

        raise ValueError("only one table can be loaded at once")

    @staticmethod
    def create_empty() -> "CatalogFrame":
        return CatalogFrame(
            {
                "namespace": [],
                "version": [],
                "table": [],
                "dimensions": [],
                "path": [],
                "format": [],
            }
        )


class CatalogSeries(pd.Series):
    _metadata = ["_base_uri"]

    @property
    def _constructor(self) -> type:
        return CatalogSeries

    def load(self) -> Table:
        if self.path and self.format and self._base_uri:
            uri = self._base_uri + self.path
            if self.format == "feather":
                return Table.read_feather(uri + ".feather")
            elif self.format == "csv":
                return Table.read_csv(uri + ".csv")
            else:
                raise ValueError("unknown format")

        raise ValueError("series is not a table spec")


def find(
    table: Optional[str] = None,
    namespace: Optional[str] = None,
    dataset: Optional[str] = None,
) -> "CatalogFrame":
    global REMOTE_CATALOG

    if not REMOTE_CATALOG:
        REMOTE_CATALOG = RemoteCatalog()

    return REMOTE_CATALOG.find(table=table, namespace=namespace, dataset=dataset)


def find_one(*args: Optional[str], **kwargs: Optional[str]) -> Table:
    return find(*args, **kwargs).load()


class PackageUpdateRequired(Exception):
    pass



#
#  variables.py
#

from os import path
import json
from typing import Any, Dict, Optional, cast

import pandas as pd

from .properties import metadata_property
from .meta import VariableMeta

SCHEMA = json.load(open(path.join(path.dirname(__file__), "schemas", "table.json")))
METADATA_FIELDS = list(SCHEMA["properties"])


class Variable(pd.Series):
    _name: Optional[str] = None
    _fields: Dict[str, VariableMeta]

    def __init__(
        self,
        data: Any = None,
        index: Any = None,
        _fields: Optional[Dict[str, VariableMeta]] = None,
        **kwargs: Any,
    ) -> None:
        self._fields = _fields or {}

        # silence warning
        if data is None and not kwargs.get("dtype"):
            kwargs["dtype"] = "object"

        super().__init__(data=data, index=index, **kwargs)

    @property
    def name(self) -> Optional[str]:
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        # move metadata when you rename a field
        if self._name and self._name in self._fields:
            self._fields[name] = self._fields.pop(self._name)

        # make sure there is always a placeholder metadata object
        if name not in self._fields:
            self._fields[name] = VariableMeta()

        self._name = name

    @property
    def checked_name(self) -> str:
        if not self.name:
            raise ValueError("variable must be named to have metadata")

        return self.name

    # which fields should pandas propagate on slicing, etc?
    _metadata = ["_fields", "_name"]

    @property
    def _constructor(self) -> type:
        return Variable

    @property
    def _constructor_expanddim(self) -> type:
        # XXX lazy circular import
        from . import tables

        return tables.Table

    @property
    def metadata(self) -> VariableMeta:
        return self._fields[self.checked_name]

    def astype(self, *args: Any, **kwargs: Any) -> "Variable":
        # To fix: https://github.com/owid/owid-catalog-py/issues/12
        v = super().astype(*args, **kwargs)
        v.name = self.name
        return cast(Variable, v)


# dynamically add all metadata properties to the class
for k in VariableMeta.__dataclass_fields__:  # type: ignore
    if hasattr(Variable, k):
        raise Exception(f'metadata field "{k}" would overwrite a Pandas built-in')

    setattr(Variable, k, metadata_property(k))



__version__ = "0.1.0"

from .catalogs import LocalCatalog, RemoteCatalog, find, find_one
from .datasets import Dataset
from .tables import Table
from .variables import Variable
from .meta import DatasetMeta, TableMeta, VariableMeta, Source, License

__all__ = [
    "LocalCatalog",
    "RemoteCatalog",
    "find",
    "find_one",
    "Dataset",
    "Table",
    "Variable",
    "DatasetMeta",
    "TableMeta",
    "VariableMeta",
    "Source",
    "License",
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['owid', 'owid.catalog']

package_data = \
{'': ['*'], 'owid.catalog': ['schemas/*']}

install_requires = \
['dataclasses-json>=0.5.6,<0.6.0',
 'ipdb>=0.13.9,<0.14.0',
 'jsonschema>=3.2.0,<4.0.0',
 'pandas-stubs>=1.2.0,<2.0.0',
 'pandas>=1.3.3,<2.0.0',
 'pyarrow>=5.0.0,<6.0.0',
 'pytest-cov>=2.12.1,<3.0.0',
 'requests>=2.26.0,<3.0.0']

setup_kwargs = {
    'name': 'owid-catalog',
    'version': '0.2.5',
    'description': 'Core data types used by OWID for managing data.',
    'long_description': '![build status](https://github.com/owid/owid-catalog-py/actions/workflows/python-package.yml/badge.svg)\n[![PyPI version](https://badge.fury.io/py/owid-catalog.svg)](https://badge.fury.io/py/owid-catalog)\n![](https://img.shields.io/badge/python-3.8%20%7C%203.9-blue.svg)\n\n# owid-catalog\n\n_A Pythonic API for working with OWID\'s data catalog._\n\nStatus: experimental, APIs likely to change\n\n## Quickstart\n\nInstall with `pip install owid-catalog`. Then you can begin exploring the experimental data catalog:\n\n```python\nfrom owid import catalog\n\n# look for Covid-19 data, return a data frame of matches\ncatalog.find(\'covid\')\n\n# load Covid-19 data from the Our World In Data namespace as a data frame\ndf = catalog.find(\'covid\', namespace=\'owid\').load()\n```\n\n## Development\n\nYou need Python 3.9+, `poetry` and `make` installed. Clone the repo, then you can simply run:\n\n```\n# run all unit tests and CI checks\nmake test\n\n# watch for changes, then run all checks\nmake watch\n```\n\n## Data types\n\n### Catalog\n\nA catalog is an arbitrarily deep folder structure containing datasets inside. It can be local on disk, or remote.\n\n#### Load the remote catalog\n\n```python\n# find the default OWID catalog and fetch the catalog index over HTTPS\ncat = RemoteCatalog()\n\n# get a list of matching tables in different datasets\nmatches = cat.find(\'population\')\n\n# fetch a data frame for a specific match over HTTPS\nt = cat.find_one(\'population\', namespace=\'gapminder\')\n```\n\n### Datasets\n\nA dataset is a folder of tables containing metadata about the overall collection.\n\n- Metadata about the dataset lives in `index.json`\n- All tables in the folder must share a common format (CSV or Feather)\n\n#### Create a new dataset\n\n```python\n# make a folder and an empty index.json file\nds = Dataset.create(\'/tmp/my_data\')\n```\n\n```python\n# choose CSV instead of feather for files\nds = Dataset.create(\'/tmp/my_data\', format=\'csv\')\n```\n\n#### Add a table to a dataset\n\n```python\n# serialize a table using the table\'s name and the dataset\'s default format (feather)\n# (e.g. /tmp/my_data/my_table.feather)\nds.add(table)\n```\n\n#### Remove a table from a dataset\n\n```python\nds.remove(\'table_name\')\n```\n\n#### Access a table\n\n```python\n# load a table including metadata into memory\nt = ds[\'my_table\']\n```\n\n#### List tables\n\n```python\n# the length is the number of datasets discovered on disk\nassert len(ds) > 0\n```\n\n```python\n# iterate over the tables discovered on disk\nfor table in ds:\n    do_something(table)\n```\n\n#### Add metadata\n\n```python\n# you need to manually save your changes\nds.title = "Very Important Dataset"\nds.description = "This dataset is a composite of blah blah blah..."\nds.save()\n```\n\n#### Copy a dataset\n\n```python\n# copying a dataset copies all its files to a new location\nds_new = ds.copy(\'/tmp/new_data_path\')\n\n# copying a dataset is identical to copying its folder, so this works too\nshutil.copytree(\'/tmp/old_data\', \'/tmp/new_data_path\')\nds_new = Dataset(\'/tmp/new_data_path\')\n```\n### Tables\n\nTables are essentially pandas DataFrames but with metadata. All operations on them occur in-memory, except for loading from and saving to disk. On disk, they are represented by tabular file (feather or CSV) and a JSON metadata file.\n\n#### Make a new table\n\n```python\n# same API as DataFrames\nt = Table({\n    \'gdp\': [1, 2, 3],\n    \'country\': [\'AU\', \'SE\', \'CH\']\n}).set_index(\'country\')\n```\n\n#### Add metadata about the whole table\n\n```python\nt.title = \'Very important data\'\n```\n\n#### Add metadata about a field\n\n```python\nt.gdp.description = \'GDP measured in 2011 international $\'\nt.sources = [\n    Source(title=\'World Bank\', url=\'https://www.worldbank.org/en/home\')\n]\n```\n\n#### Add metadata about all fields at once\n\n```python\n# sources and licenses are actually stored a the field level\nt.sources = [\n    Source(title=\'World Bank\', url=\'https://www.worldbank.org/en/home\')\n]\nt.licenses = [\n    License(\'CC-BY-SA-4.0\', url=\'https://creativecommons.org/licenses/by-nc/4.0/\')\n]\n```\n\n#### Save a table to disk\n\n```python\n# save to /tmp/my_table.feather + /tmp/my_table.meta.json\nt.to_feather(\'/tmp/my_table.feather\')\n\n# save to /tmp/my_table.csv + /tmp/my_table.meta.json\nt.to_csv(\'/tmp/my_table.csv\')\n```\n\n#### Load a table from disk\n\nThese work like normal pandas DataFrames, but if there is also a `my_table.meta.json` file, then metadata will also get read. Otherwise it will be assumed that the data has no metadata:\n\n```python\nt = Table.read_feather(\'/tmp/my_table.feather\')\n\nt = Table.read_csv(\'/tmp/my_table.csv\')\n```\n\n\n## Changelog\n\n- `v0.2.5`\n    - Fix ability to load remote CSV tables\n- `v0.2.4`\n    - Update the default catalog URL to use a CDN\n- `v0.2.3`\n    - Fix methods for finding and loading data from a `LocalCatalog`\n- `v0.2.2`\n    - Repack frames to compact dtypes on `Table.to_feather()`\n- `v0.2.1`\n    - Fix key typo used in version check\n- `v0.2.0`\n    - Copy dataset metadata into tables, to make tables more traceable\n    - Add API versioning, and a requirement to update if your version of this library is too old\n- `v0.1.1`\n    - Add support for Python 3.8\n- `v0.1.0`\n    - Initial release, including searching and fetching data from a remote catalog\n',
    'author': 'Our World In Data',
    'author_email': 'tech@ourworldindata.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/owid/owid-grapher-py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



#
#  datasets.py
#

from os.path import join, exists
from os import mkdir
from dataclasses import dataclass
import shutil
from typing import Any, Iterator, List, Literal, Optional, Union
from glob import glob
import hashlib
from pathlib import Path
import json

import pandas as pd

from . import tables
from .properties import metadata_property
from .meta import DatasetMeta, TableMeta


@dataclass
class Dataset:
    """
    A dataset is a folder full of data tables, with metadata available at `index.json`.
    """

    path: str
    metadata: "DatasetMeta"

    def __init__(self, path: Union[str, Path]) -> None:
        # for convenience, accept Path objects directly
        if isinstance(path, Path):
            self.path = path.as_posix()
        else:
            self.path = path

        self.metadata = DatasetMeta.load(self._index_file)

    @classmethod
    def create_empty(
        cls, path: Union[str, Path], metadata: Optional["DatasetMeta"] = None
    ) -> "Dataset":
        path = Path(path)

        if path.is_dir():
            if not (path / "index.json").exists():
                raise Exception(f"refuse to overwrite non-dataset dir at: {path}")
            shutil.rmtree(path)

        mkdir(path)

        index_file = path / "index.json"
        DatasetMeta().save(index_file)

        return Dataset(path.as_posix())

    def add(
        self, table: tables.Table, format: Literal["csv", "feather"] = "feather"
    ) -> None:
        """Add this table to the dataset by saving it in the dataset's folder. Defaults to
        feather format but you can override this to csv by passing 'csv' for the format"""

        # copy dataset metadata to the table
        table.metadata.dataset = self.metadata

        allowed_formats = ["feather", "csv"]
        if format not in allowed_formats:
            raise Exception(f"Format '{format}'' is not supported")
        table_filename = join(self.path, table.metadata.checked_name + f".{format}")
        if format == "feather":
            table.to_feather(table_filename)
        else:
            table.to_csv(table_filename)

    def __getitem__(self, name: str) -> tables.Table:
        table_filename = join(self.path, name + ".feather")
        if exists(table_filename):
            return tables.Table.read_feather(table_filename)
        table_filename = join(self.path, name + ".csv")
        if exists(table_filename):
            return tables.Table.read_csv(table_filename)
        raise KeyError(name)

    def __contains__(self, name: str) -> bool:
        feather_table_filename = join(self.path, name + ".feather")
        csv_table_filename = join(self.path, name + ".csv")
        return exists(feather_table_filename) or exists(csv_table_filename)

    def save(self) -> None:
        self.metadata.save(self._index_file)
        self._update_table_metadata()

    def _update_table_metadata(self) -> None:
        "Update the copy of this dataset's metadata in every table in the set."
        dataset_meta = self.metadata.to_dict()

        for metadata_file in glob(join(self.path, "*.meta.json")):
            with open(metadata_file) as istream:
                table_meta = json.load(istream)

            table_meta["dataset"] = dataset_meta

            with open(metadata_file, "w") as ostream:
                json.dump(table_meta, ostream, indent=2)

    def index(self, catalog_path: Path) -> pd.DataFrame:
        """
        Return a DataFrame describing the contents of this dataset, one row per table.
        """
        base = {
            "namespace": self.metadata.namespace,
            "dataset": self.metadata.short_name,
            "version": self.metadata.version,
            "checksum": self.checksum(),
        }
        rows = []
        for metadata_file in self._metadata_files:
            with open(metadata_file) as istream:
                metadata = TableMeta.from_dict(json.load(istream))

            row = base.copy()

            assert metadata.short_name
            row["table"] = metadata.short_name

            row["dimensions"] = json.dumps(metadata.primary_key)

            table_path = Path(self.path) / metadata.short_name
            row["path"] = table_path.relative_to(catalog_path).as_posix()

            if table_path.with_suffix(".feather").exists():
                row["format"] = "feather"
            elif table_path.with_suffix(".csv").exists():
                row["format"] = "csv"

            rows.append(row)

        return pd.DataFrame.from_records(rows)

    @property
    def _index_file(self) -> str:
        return join(self.path, "index.json")

    def __len__(self) -> int:
        return len(self._data_files)

    def __iter__(self) -> Iterator[tables.Table]:
        for filename in self._data_files:
            if filename.endswith(".feather"):
                yield tables.Table.read_feather(filename)

            elif filename.endswith(".csv"):
                yield tables.Table.read_csv(filename)

            else:
                raise Exception(f"don't know how to read table: {filename}")

    @property
    def _data_files(self) -> List[str]:
        feather_pattern = join(self.path, "*.feather")
        csv_pattern = join(self.path, "*.csv")
        return sorted(glob(feather_pattern) + glob(csv_pattern))

    @property
    def _metadata_files(self) -> List[str]:
        return sorted(glob(join(self.path, "*.meta.json")))

    def checksum(self) -> str:
        "Return a MD5 checksum of all data and metadata in the dataset."
        _hash = hashlib.md5()
        _hash.update(checksum_file(self._index_file).digest())

        for data_file in self._data_files:
            _hash.update(checksum_file(data_file).digest())

            metadata_file = Path(data_file).with_suffix(".meta.json").as_posix()
            _hash.update(checksum_file(metadata_file).digest())

        return _hash.hexdigest()


for k in DatasetMeta.__dataclass_fields__:  # type: ignore
    if hasattr(Dataset, k):
        raise Exception(f'metadata field "{k}" would overwrite a Dataset built-in')

    setattr(Dataset, k, metadata_property(k))


def checksum_file(filename: str) -> Any:
    "Return the MD5 checksum of a given file."
    chunk_size = 2 ** 20  # 1MB
    checksum = hashlib.md5()
    with open(filename, "rb") as istream:
        chunk = istream.read(chunk_size)
        while chunk:
            checksum.update(chunk)
            chunk = istream.read(chunk_size)

    return checksum
